# Generated by Django 5.0.1 on 2024-02-09 22:24

import django.db.models.deletion
import modelcluster.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='FilterSettings',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_updated', models.DateTimeField(auto_now=True, verbose_name='Last Updated')),
            ],
            options={
                'verbose_name': 'Request Filter Settings',
                'verbose_name_plural': 'Request Filter Settings',
            },
        ),
        migrations.CreateModel(
            name='Filter',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sort_order', models.IntegerField(blank=True, editable=False, null=True)),
                ('filter_type', models.CharField(choices=[('IP', 'IP'), ('USER_AGENT', 'User Agent'), ('PATH', 'Path'), ('QUERY_STRING', 'Query String'), ('REFERER', 'Referer'), ('COUNTRY', 'Country')], default='IP', max_length=20, verbose_name='Filter Type')),
                ('method', models.CharField(choices=[('ABSOLUTE', 'Absolute'), ('WILDCARD', 'Wildcard'), ('REGEX', 'Regex')], default='ABSOLUTE', max_length=20, verbose_name='Method')),
                ('action', models.CharField(choices=[('ALLOW', 'Allow'), ('BLOCK', 'Block'), ('REDIRECT', 'Redirect'), ('LOG', 'Log')], default='BLOCK', max_length=20, verbose_name='Action')),
                ('action_value', models.CharField(blank=True, help_text='The value to use for the action. For example, if you are redirecting, you would enter the URL here.', max_length=255, verbose_name='Action Value')),
                ('filter_value', models.CharField(help_text='The value to filter on. For example, if you are filtering on IP, you would enter an IP address or CIDR range here.', max_length=255, verbose_name='Filter Value')),
                ('active', models.BooleanField(default=True, verbose_name='Active')),
                ('settings', modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.CASCADE, related_name='filters', to='request_filters.filtersettings')),
            ],
            options={
                'verbose_name': 'Filter',
                'verbose_name_plural': 'Filters',
            },
        ),
    ]



import logging
from django.db import models
from django.http import HttpRequest, HttpResponse
from django.conf import settings as django_settings
from django.utils.translation import (
    gettext_lazy as _,
    get_language_from_request,
)

from geoip2.errors import AddressNotFoundError
from django.contrib.gis.geoip2 import (
    GeoIP2Exception,
)

from .filters import Filter
from ..util import (
    get_ip_address,
    get_country,
    logging_skipped,
    skip_logging,
)

from .. import log


class LogFailed(Exception):
    pass


class FilteredRequest(models.Model):
    """
        A FilteredRequest is a request that has failed a filter check.
    """

    _filter = models.JSONField(
        verbose_name=_("Filter"),
        help_text=_("The filter that was triggered."),
        blank=True,
        null=False,
        default=dict,
    )

    request_path = models.CharField(
        max_length=255,
        verbose_name=_("Request Path"),
        help_text=_("The path of the request that was filtered."),
    )

    request_method = models.CharField(
        max_length=10,
        verbose_name=_("Request Method"),
        help_text=_("The method of the request that was filtered."),
    )

    request_ip = models.GenericIPAddressField(
        verbose_name=_("Request IP"),
        help_text=_("The IP of the request that was filtered."),
    )

    request_language = models.CharField(
        max_length=10,
        verbose_name=_("Request Language"),
        help_text=_("The language of the request that was filtered."),
    )

    request_is_secure = models.BooleanField(
        verbose_name=_("Request Is Secure"),
        help_text=_("Whether the request was secure."),
    )

    response_info = models.JSONField(
        verbose_name=_("Response Info"),
        help_text=_("Extra information about the response after the request was filtered."),
        blank=True,
        null=False,
        default=dict,
    )

    request_info = models.JSONField(
        verbose_name=_("Request Info"),
        help_text=_("Extra information about the request that was filtered."),
        blank=True,
        null=False,
        default=dict,
    )

    gis_data = models.JSONField(
        verbose_name=_("GIS Data"),
        help_text=_("The GIS data of the request that was filtered."),
        blank=True,
        null=False,
        default=dict,
    )

    created_at = models.DateTimeField(
        auto_now_add=True,
        verbose_name=_("Created At"),
    )

    class Meta:
        verbose_name = _("Filtered Request")
        verbose_name_plural = _("Filtered Requests")
        ordering = ["-created_at"]


    def __str__(self):

        s = []

        if self._filter:
            s.append(self.filter.get_action_display())

        s.append(self.request_ip)
        s.append(self.request_method)

        if self.request_language:
            s.append(self.request_language)

        if self.gis_data:
            n = self.gis_data.get("country_name", "")
            if n: s.append(n)

        if not self.pk:
            return f"FilteredRequest({'/'.join(s)})"

        return f"FilteredRequest({self.pk}: {'/'.join(s)})"
    
    @property
    def filter(self):
        return Filter(**self._filter)
    
    @classmethod
    def skip_logging(cls, request_or_response):
        """
            Note that the log action should stop if the req/res is skipped
        """
        return skip_logging(request_or_response)
    
    @classmethod
    def should_log(cls, request_or_response):
        """
            Check if the request or response should be logged
        """
        return not logging_skipped(request_or_response)

    @classmethod
    def log_request(cls, request: HttpRequest, filter: "Filter", response: HttpResponse = None, fail_silently: bool = False, commit: bool = True):
        """
            Log a request that has been filtered
        """

        if not cls.should_log(request):
            log(f"({cls.__name__}) Request {request} has been skipped from logging.", level=logging.DEBUG)
            return
        
        if response and not cls.should_log(response):
            log(f"({cls.__name__}) Response {response} has been skipped from logging.", level=logging.DEBUG)
            return

        ip = get_ip_address(request)
        if not ip:
            log(f"({cls.__name__}) Could not get the IP address from the request ([{request.method}] - {request.path}).", level=logging.ERROR)
            if fail_silently: return
            raise LogFailed(f"Could not get the IP address from the request.")

        try:
            country = get_country(ip)
        except (TypeError, ValueError, GeoIP2Exception, AddressNotFoundError):
            log(f"({cls.__name__}) Could not get the country information for IP address ([{ip}/{request.method}] - {request.path}).", level=logging.ERROR)
            country = {}
        
        data = {
            "GET":          request.GET,
            "POST":         request.POST,
            "HAS_FILES":    bool(request.FILES),
            "HEADERS":      dict(request.headers),
        }

        _create_kwargs = {}

        if filter:
            if not filter.pk:
                err = f"Filter {filter} has not been saved for ([{ip}/{request.method}] - {request.path}])."
                log(f"({cls.__name__}) {err}", level=logging.ERROR)
                if fail_silently: return
                raise LogFailed(err)

            _create_kwargs["_filter"] = {
                "pk":           filter.pk,
                "settings_id":  filter.settings_id,
                "filter_type":  filter.filter_type,
                "method":       filter.method,
                "action":       filter.action,
                "action_value": filter.action_value,
                "filter_value": filter.filter_value,
                "active":       filter.active,
            }

        if response:
            _create_kwargs["response_info"] = {
                "STATUS_CODE": response.status_code,
                "HEADERS":     dict(response.items()),
            }

        if django_settings.USE_I18N:
            _create_kwargs["request_language"] = get_language_from_request(request)

        self = cls(
            gis_data            = country,
            request_ip          = ip,
            request_path        = request.path,
            request_method      = request.method,
            request_is_secure   = request.is_secure(),
            request_info        = data,
            **_create_kwargs,
        )

        if commit:
            log(f"({cls.__name__}) Logging filtered request {self} with {filter} ([{ip}/{request.method}] - {request.path}).", level=logging.DEBUG)
            self.save()

        return self
    



# Generated by Django 5.0.1 on 2024-02-10 08:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('request_filters', '0003_filteredrequest'),
    ]

    operations = [
        migrations.AddField(
            model_name='filteredrequest',
            name='response_info',
            field=models.JSONField(blank=True, default=dict, help_text='Extra information about the response after the request was filtered.', verbose_name='Response Info'),
        ),
    ]



# Generated by Django 5.0.1 on 2024-02-10 08:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('request_filters', '0002_alter_filter_options_alter_filter_filter_type_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='FilteredRequest',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('_filter', models.JSONField(blank=True, default=dict, help_text='The filter that was triggered.', verbose_name='Filter')),
                ('request_path', models.CharField(help_text='The path of the request that was filtered.', max_length=255, verbose_name='Request Path')),
                ('request_method', models.CharField(help_text='The method of the request that was filtered.', max_length=10, verbose_name='Request Method')),
                ('request_ip', models.GenericIPAddressField(help_text='The IP of the request that was filtered.', verbose_name='Request IP')),
                ('request_language', models.CharField(help_text='The language of the request that was filtered.', max_length=10, verbose_name='Request Language')),
                ('request_is_secure', models.BooleanField(help_text='Whether the request was secure.', verbose_name='Request Is Secure')),
                ('request_info', models.JSONField(blank=True, default=dict, help_text='Extra information about the request that was filtered.', verbose_name='Request Info')),
                ('gis_data', models.JSONField(blank=True, default=dict, help_text='The GIS data of the request that was filtered.', verbose_name='GIS Data')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created At')),
            ],
            options={
                'verbose_name': 'Filtered Request',
                'verbose_name_plural': 'Filtered Requests',
                'ordering': ['-created_at'],
            },
        ),
    ]



from django.db import models
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _
from django.contrib.gis.geoip2 import (
    GeoIP2Exception,
)
from .options import (
    RequestFilters,
)
from .util import (
    get_ip_address,
    ip_in_cidr,
    get_country,
)
from typing import TYPE_CHECKING, Callable
import re, fnmatch

if TYPE_CHECKING:
    from .models import FilterSettings, Filter

# yeah naming kinda sucks.
_py_filter = filter

class FilterChoices(models.TextChoices):
    IP              = "IP",             _("IP")
    USER_AGENT      = "USER_AGENT",     _("User Agent")
    PATH            = "PATH",           _("Path")
    QUERY_STRING    = "QUERY_STRING",   _("Query String")
    REFERER         = "REFERER",        _("Referer")
    COUNTRY         = "COUNTRY",        _("Country")
    METHOD          = "METHOD",         _("Method")

class FilterMethodChoices(models.TextChoices):
    ABSOLUTE = "ABSOLUTE",  _("Absolute")
    WILDCARD = "WILDCARD",  _("Wildcard")
    REGEX    = "REGEX",     _("Regex")
    IN       = "IN",        _("In")


"""
    Absolute checks
"""

def ABSOLUTE_FILTER_CHECK_IP(filter: "Filter", settings: "FilterSettings", request: HttpRequest):
    ip = get_ip_address(request)
    try:
        return ip_in_cidr(ip, filter.filter_value)
    except ValueError:
        return RequestFilters.default_check_value(filter, settings, request)
    
def ABSOLUTE_FILTER_CHECK_USER_AGENT(filter: "Filter", settings: "FilterSettings", request: HttpRequest):
    # The only absolute check that checks for inclusion.
    # The user agent can wildly vary.
    return filter.filter_value in request.META.get("HTTP_USER_AGENT", "") or\
            filter.filter_value == request.META.get("HTTP_USER_AGENT", "")

def ABSOLUTE_FILTER_CHECK_PATH(filter: "Filter", settings: "FilterSettings", request: HttpRequest):
    return request.path.lower() == filter.filter_value.lower()

def ABSOLUTE_FILTER_CHECK_QUERY_STRING(filter: "Filter", settings: "FilterSettings", request: HttpRequest):
    return filter.action_value == request.GET.get(filter.filter_value, "")

def ABSOLUTE_FILTER_CHECK_REFERER(filter: "Filter", settings: "FilterSettings", request: HttpRequest):
    return filter.filter_value == request.META.get("HTTP_REFERER", "")

def ABSOLUTE_FILTER_CHECK_COUNTRY(filter: "Filter", settings: "FilterSettings", request: HttpRequest):
    ip = get_ip_address(request)
    try:
        country = get_country(ip)
    except (TypeError, ValueError, GeoIP2Exception):
        return RequestFilters.default_check_value(filter, settings, request)
    
    code = country.get("country_code", "")
    name = country.get("country_name", "")

    return filter.filter_value in (code, name)

def ABSOLUTE_FILTER_CHECK_METHOD(filter: "Filter", settings: "FilterSettings", request: HttpRequest):
    return filter.filter_value == request.method

"""
    Contains checks
"""

IN_FILTER_CHECK_IP = ABSOLUTE_FILTER_CHECK_IP


def IN_FILTER_CHECK_QUERY_STRING(filter: "Filter", settings: "FilterSettings", request: HttpRequest):
    values = list(_py_filter(lambda x: x.strip(), filter.filter_value.split(",")))
    actions = list(_py_filter(lambda x: x.strip(), filter.action_value.split(",")))
    if len(values) != len(actions):
        for value in values:
            if value in request.GET and request.GET.get(value, "") in actions:
                return True
        return False
    
    return any(request.GET.get(value, "") in actions for value in values)

def IN_FILTER_CHECK_COUNTRY(filter: "Filter", settings: "FilterSettings", request: HttpRequest):
    ip = get_ip_address(request)
    try:
        country = get_country(ip)
    except (TypeError, ValueError, GeoIP2Exception):
        return RequestFilters.default_check_value(filter, settings, request)
    
    code = country.get("country_code", "")
    name = country.get("country_name", "")

    return any(filter_value in (code, name) for filter_value in filter.filter_value.split(","))

def IN_FILTER_CHECK_METHOD(filter: "Filter", settings: "FilterSettings", request: HttpRequest):
    return request.method in filter.filter_value.split(",")

"""
    Regular expression checks
"""

def CHECK_GET_IP_ADDRESS(filter: "Filter", settings: "FilterSettings", request: HttpRequest):
    return get_ip_address(request)

def CHECK_GET_USER_AGENT(filter: "Filter", settings: "FilterSettings", request: HttpRequest):
    return request.META.get("HTTP_USER_AGENT", "")

def CHECK_GET_PATH(filter: "Filter", settings: "FilterSettings", request: HttpRequest):
    return request.path

def CHECK_GET_QUERY_STRING(filter: "Filter", settings: "FilterSettings", request: HttpRequest):
    return request.GET.get(filter.filter_value, "")

def CHECK_GET_REFERER(filter: "Filter", settings: "FilterSettings", request: HttpRequest):
    return request.META.get("HTTP_REFERER", "")

def CHECK_GET_COUNTRY(filter: "Filter", settings: "FilterSettings", request: HttpRequest):
    ip = get_ip_address(request)
    try:
        country = get_country(ip)
    except (TypeError, ValueError, GeoIP2Exception):
        raise ValueError("Could not get country information for IP address")
    
    return country.get("country_code", "")

def CHECK_GET_METHOD(filter: "Filter", settings: "FilterSettings", request: HttpRequest):
    return request.method


def RE_MATCH_CHECK(get_value: Callable[["Filter", "FilterSettings", HttpRequest], str]):
    def _inner(filter: "Filter", settings: "FilterSettings", request: HttpRequest):
        value = get_value(filter, settings, request)
        match = re.match(filter.filter_value, value)
        return bool(match)
    return _inner


def WC_MATCH_CHECK(get_value: Callable[["Filter", "FilterSettings", HttpRequest], str]):
    def _inner(filter: "Filter", settings: "FilterSettings", request: HttpRequest):
        value = get_value(filter, settings, request)
        return fnmatch.fnmatch(value, filter.filter_value)
    return _inner


def IN_MATCH_CHECK(get_value: Callable[["Filter", "FilterSettings", HttpRequest], str], *process: Callable[[str], str], split_by = ","):
    def _inner(filter: "Filter", settings: "FilterSettings", request: HttpRequest):
        value = get_value(filter, settings, request)
        if isinstance(value, str):
            value = [value]

        if not filter.filter_value:
            # Special case.
            return False

        filter_value = filter.filter_value
        for p in process:
            filter_value = p(filter_value)

        if not split_by:
            return filter_value in value

        split = _s_str(filter_value, split_by)
        for val in value:
            if any(v in val for v in split):
                return True
            
        return False

    return _inner

def _p_upper(s):
    return s.upper()

def _s_str(s, split_by = ","):
    if not s:
        return [s]
    return list(_py_filter(lambda x: x.strip(), s.split(split_by)))

# Filters that are absolute and do not require any additional checks
FILTER_CHECKS_ABSOLUTE = {
    FilterChoices.IP:              ABSOLUTE_FILTER_CHECK_IP,
    FilterChoices.USER_AGENT:      ABSOLUTE_FILTER_CHECK_USER_AGENT,
    FilterChoices.PATH:            ABSOLUTE_FILTER_CHECK_PATH,
    FilterChoices.QUERY_STRING:    ABSOLUTE_FILTER_CHECK_QUERY_STRING,
    FilterChoices.REFERER:         ABSOLUTE_FILTER_CHECK_REFERER,
    FilterChoices.COUNTRY:         ABSOLUTE_FILTER_CHECK_COUNTRY,
    FilterChoices.METHOD:          ABSOLUTE_FILTER_CHECK_METHOD,
}

FILTER_CHECKS_RE = {
    FilterChoices.IP:              RE_MATCH_CHECK(CHECK_GET_IP_ADDRESS),
    FilterChoices.USER_AGENT:      RE_MATCH_CHECK(CHECK_GET_USER_AGENT),
    FilterChoices.PATH:            RE_MATCH_CHECK(CHECK_GET_PATH),
    FilterChoices.QUERY_STRING:    RE_MATCH_CHECK(CHECK_GET_QUERY_STRING),
    FilterChoices.REFERER:         RE_MATCH_CHECK(CHECK_GET_REFERER),
    FilterChoices.COUNTRY:         RE_MATCH_CHECK(CHECK_GET_COUNTRY),
    FilterChoices.METHOD:          RE_MATCH_CHECK(CHECK_GET_METHOD),
}


FILTER_CHECKS_WC = {
    FilterChoices.IP:              WC_MATCH_CHECK(CHECK_GET_IP_ADDRESS),
    FilterChoices.USER_AGENT:      WC_MATCH_CHECK(CHECK_GET_USER_AGENT),
    FilterChoices.PATH:            WC_MATCH_CHECK(CHECK_GET_PATH),
    FilterChoices.QUERY_STRING:    WC_MATCH_CHECK(CHECK_GET_QUERY_STRING),
    FilterChoices.REFERER:         WC_MATCH_CHECK(CHECK_GET_REFERER),
    FilterChoices.COUNTRY:         WC_MATCH_CHECK(CHECK_GET_COUNTRY),
    FilterChoices.METHOD:          WC_MATCH_CHECK(CHECK_GET_METHOD),
}

FILTER_CHECKS_IN = {
    FilterChoices.IP:              IN_FILTER_CHECK_IP,
    FilterChoices.USER_AGENT:      IN_MATCH_CHECK(CHECK_GET_USER_AGENT, split_by = None),
    FilterChoices.PATH:            IN_MATCH_CHECK(CHECK_GET_PATH, split_by = None),
    FilterChoices.QUERY_STRING:    IN_FILTER_CHECK_QUERY_STRING,
    FilterChoices.REFERER:         IN_MATCH_CHECK(CHECK_GET_REFERER, split_by = None),
    FilterChoices.COUNTRY:         IN_FILTER_CHECK_COUNTRY,
    FilterChoices.METHOD:          IN_MATCH_CHECK(CHECK_GET_METHOD, _p_upper, split_by = " "),
}

FILTER_CHECKS = {
    FilterMethodChoices.ABSOLUTE: FILTER_CHECKS_ABSOLUTE,
    FilterMethodChoices.IN:       FILTER_CHECKS_IN,
    FilterMethodChoices.REGEX:    FILTER_CHECKS_RE,
    FilterMethodChoices.WILDCARD: FILTER_CHECKS_WC,
}





default_app_config = "request_filters.apps.RequestFiltersConfig"

import logging as _logging
_logger = _logging.getLogger("request_filters")

# LOG_FORMAT = "%(asctime)s [%(levelname)s / %(name)s]: %(message)s"
def log(message, level=_logging.INFO, *args, **kwargs):
    s = f'[{_logging.getLevelName(level)} / {_logger.name}]: {message}'
    _logger.log(level, s, *args, **kwargs)

log.DEBUG = _logging.DEBUG
log.INFO = _logging.INFO
log.WARN = _logging.WARN
log.ERROR = _logging.ERROR
log.CRITICAL = _logging.CRITICAL

VERSION = (1, 0, 0)
__version__ = ".".join(map(str, VERSION))






from django.test import TestCase
import ipaddress, random

from .models import (
    Filter,
    FilterChoices,
    FilterMethodChoices,
    FilterActionChoices,
)
from .util import ip_in_cidr


def generate_ips_for_test(cidr, inside_count, outside_count):
    network = ipaddress.ip_network(cidr)
    all_ips = list(network.hosts())

    # Select inside IPs
    inside_ips = random.sample(all_ips, min(inside_count, len(all_ips)))

    # Generate outside IPs
    outside_ips = []
    while len(outside_ips) < outside_count:
        random_ip = ipaddress.IPv4Address(random.randint(0, 2**32 - 1))
        if random_ip not in network:
            outside_ips.append(str(random_ip))

    # Prepare the dictionary for test cases
    test_ips = {
        'inside': [str(ip) for ip in inside_ips],
        'outside': outside_ips,
    }
    
    return test_ips

# Example CIDR ranges for testing
ranges = [
    "222.240.205.0/24",
    "185.76.203.0/24",
    "46.227.139.0/24",
    "21.133.158.0/24",
    "155.173.115.0/24",
]

user_agents = [
    "Googlebot",
    "Bingbot",
    "Yandexbot",
    "DuckDuckBot",
    "Baiduspider",
]

fail_agents = [
    "test",
    "debug",
    "password",
    "admin",
    "login",
    "logout",
]

paths = [
    "/admin",
    "/login",
    "/logout",
    "/register",
    "/password_reset",
    "/password_change",
    "/profile",
    "/settings",
]

fail_paths = [
    "/fail/admin",
    "/fail/login",
    "/fail/logout",
    "/fail/register",
    "/fail/password_reset",
    "/fail/password_change",
    "/fail/profile",
    "/fail/settings",
]

query_strings = [
    "test",
    "debug",
    "password",
    "admin",
    "login",
    "logout",
]

fail_query_strings = [
    "fail",
    "error",
    "exception",
    "warning",
    "critical",
]



class TestIPAddr(TestCase):
    def test_ip_in_cidr(self):
        for cidr in ranges:
            ips = generate_ips_for_test(cidr, 10, 10)
            for ip in ips['inside']:
                self.assertTrue(ip_in_cidr(ip, cidr))

            for ip in ips['outside']:
                self.assertFalse(ip_in_cidr(ip, cidr))


def get_response(request):
    return "SUCCESS_RESPONSE"

# from .middleware import RequestFilterMiddleware
# 
# class TestRequestFilterMiddleware(TestCase):
#     def test_request_filter_middleware(self):
#         middleware = RequestFilterMiddleware(get_response)
#         request = type("Request", (), {"META": {}})()
#         response = middleware(request)
#         self.assertEqual(response, "SUCCESS_RESPONSE")



filters = []

for cidr in ranges:
    filters.append(Filter(
        filter_type=FilterChoices.IP,
        method=FilterMethodChoices.ABSOLUTE,
        action=FilterActionChoices.BLOCK,
        filter_value=cidr,
    ))

filters += [
    Filter(
        filter_type=FilterChoices.IP,
        method=FilterMethodChoices.REGEX,
        action=FilterActionChoices.BLOCK,
        filter_value="\d+\.\d+\.\d+\.\d+",
    ),
]


# filters += Filter(
#     filter_type=FilterChoices.USER_AGENT,
#     method=FilterMethodChoices.ABSOLUTE,
#     action=FilterActionChoices.BLOCK,
#     filter_value="Googlebot",
# )
# 
# filters += Filter(
#     filter_type=FilterChoices.PATH,
#     method=FilterMethodChoices.ABSOLUTE,
#     action=FilterActionChoices.BLOCK,
#     filter_value="/admin",
# )
# 
# filters += Filter(
#     filter_type=FilterChoices.QUERY_STRING,
#     method=FilterMethodChoices.ABSOLUTE,
#     action=FilterActionChoices.BLOCK,
#     filter_value="test",
# )


class FakeRequest:
    META = {}
    path = "/test"
    GET = {}
    POST = {}
    COOKIES = {}
    FILES = {}

    def __init__(self, path=None, query_string=None, user_agent=None, ip=None):
        self.META['REMOTE_ADDR'] = ip
        self.META['HTTP_USER_AGENT'] = user_agent
        self.path = path
        self.GET = {"test": query_string}

    def __str__(self):
        return f"FakeRequest({self.META['REMOTE_ADDR']})"

def generate_random_ips(count):
    ips = []
    for _ in range(count):
        ip = ipaddress.IPv4Address(random.randint(0, 2**32 - 1))
        ips.append(str(ip))
    return ips

class TestFilter(TestCase):
    def test_filter(self):
        for filter in filters:
            filter: Filter
            if filter.filter_type == FilterChoices.IP:
                if filter.method == FilterMethodChoices.ABSOLUTE:
                    ips = generate_ips_for_test(filter.filter_value, 10, 10)

                    for ip in ips['inside']:
                        self.assertTrue(filter.passes_test(None, FakeRequest(ip=ip)), f"IP {ip} should pass for filter {filter.filter_value} ({filter.method})")

                    for ip in ips['outside']:
                        self.assertFalse(filter.passes_test(None, FakeRequest(ip=ip)), f"IP {ip} should not pass for filter {filter.filter_value} ({filter.method})")

                elif filter.method == FilterMethodChoices.REGEX:
                    ips = generate_random_ips(255 * 255)
                    for ip in ips:
                        self.assertTrue(filter.passes_test(None, FakeRequest(ip=ip)), f"IP {ip} should pass for filter {filter.filter_value} ({filter.method})")




# Generated by Django 5.0.1 on 2024-02-10 06:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('request_filters', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='filter',
            options={'ordering': ['sort_order'], 'verbose_name': 'Filter', 'verbose_name_plural': 'Filters'},
        ),
        migrations.AlterField(
            model_name='filter',
            name='filter_type',
            field=models.CharField(choices=[('IP', 'IP'), ('USER_AGENT', 'User Agent'), ('PATH', 'Path'), ('QUERY_STRING', 'Query String'), ('REFERER', 'Referer'), ('COUNTRY', 'Country'), ('METHOD', 'Method')], default='IP', max_length=20, verbose_name='Filter Type'),
        ),
        migrations.AlterField(
            model_name='filter',
            name='method',
            field=models.CharField(choices=[('ABSOLUTE', 'Absolute'), ('WILDCARD', 'Wildcard'), ('REGEX', 'Regex'), ('IN', 'In')], default='ABSOLUTE', max_length=20, verbose_name='Method'),
        ),
    ]



from django.db import models
from django.utils.translation import gettext_lazy as _
from django.apps import apps

from wagtail.contrib.settings.models import (
    BaseGenericSetting, 
    register_setting,
)
from wagtail.admin.panels import (
    HelpPanel,
    FieldPanel,
    FieldRowPanel,
    InlinePanel,
)
from modelcluster.models import ClusterableModel

from ..options import (
    RequestFilters,
)

from typing import TYPE_CHECKING
import urllib.parse

if TYPE_CHECKING:
    from .filters import Filter, FilterQuerySet

def _p(text: str, *args: str, end_line: bool = True) -> str:
    p = [
        text,
    ]
    p.extend(args)
    if end_line:
        p.append("<br/>")
    return " ".join(p)

def _a(href: str, text: str) -> str:
    if not text:
        parsed = urllib.parse.urlparse(href)
        text = parsed.netloc

    return f'<a href="{href}" target="_blank">{text}</a>'

def _l(texts: list[str], tag: str = "ol", sub_tag: str = "li") -> str:
    tags = "".join([f"<{sub_tag}>{text}</{sub_tag}>" for text in texts])
    return f"<{tag}>{tags}</{tag}>"



FilterModel: "Filter" = apps.get_model(
    "request_filters",
    "Filter",
    require_ready=False,
)


@register_setting(icon="key")
class FilterSettings(ClusterableModel, BaseGenericSetting):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._filters = []

    filters: "FilterQuerySet"

    last_updated = models.DateTimeField(
        auto_now=True,
        verbose_name=_("Last Updated"),
        editable=False,
    )

    panels = [
        HelpPanel(_("".join([
            _p("This settings page allows you to configure filters for requests."),
            _p("Filters are used for:"),
            _l([
                "Blocking requests",
                "Allowing requests",
                "Redirecting requests",
                "Logging requests",
            ], "ol"),
            _p("Learn more about and test your ", _a("https://regexr.com/", "regex")),
            _p("Learn more about", _a("https://en.wikipedia.org/wiki/Glob_(programming)", "glob patterns")),
            _p("Learn about", _a("https://en.wikipedia.org/wiki/Classless_Inter-Domain_Routing", "CIDR notation")),
        ]))),
        FieldPanel("last_updated", read_only=True),
        InlinePanel("filters", panels=[
            FieldRowPanel([
                    FieldPanel("filter_type"),
                    FieldPanel("method"),
                    FieldPanel("action"),
                    FieldPanel("active"),
                ]),
                FieldPanel("action_value"),
                FieldPanel("filter_value"),
            ], 
            label=_("Filters")
        ),
    ]

    class Meta:
        verbose_name = _("Request Filter Settings")
        verbose_name_plural = _("Request Filter Settings")

    def __str__(self):
        return f"{str(self._meta.verbose_name)} / {self.last_updated}"

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        if RequestFilters.CLEAR_CACHE_ON_SAVE_SETTINGS:
            RequestFilters.cache_backend.delete(RequestFilters.SETTINGS_CACHE_KEY)
            RequestFilters.cache_backend.delete(RequestFilters.FILTERS_CACHE_KEY)

    def get_filters(self) -> list["Filter"]:
        if self._filters is None:
            self._filters = self.filters.all()
        return self._filters
    
    @classmethod
    def base_queryset(cls):
        return super()\
            .base_queryset()\
            .prefetch_related(models.Prefetch(
                "filters", 
                queryset=FilterModel.objects\
                    .active()\
                    .order_by("sort_order"),
            ))

    @classmethod
    def load_from_cache(cls, request_or_site=None):
        """
            Load the settings from the cache, 
            or fetch it from the database and cache it
        """

        attr_name = cls.get_cache_attr_name()
        if hasattr(request_or_site, attr_name):
            return getattr(request_or_site, attr_name)

        # Try to get the settings from the cache
        self = RequestFilters.cache_backend.get(RequestFilters.SETTINGS_CACHE_KEY, None)

        if self is None:
            # Load the settings from the database
            self = cls.load(request_or_site=request_or_site)
            
            # Cache the filters for the settings
            self._filters = self.filters.all()
            
            # Cache the settings
            RequestFilters.cache_backend.set(
                RequestFilters.SETTINGS_CACHE_KEY,
                self,
                RequestFilters.SETTINGS_CACHE_TIMEOUT.total_seconds(),
            )

        # Cache for next time.
        # This gets done in the load method, but does not include the filters
        setattr(request_or_site, attr_name, self)

        return self
    






from django.apps import apps
from django.core.exceptions import PermissionDenied
from typing import TYPE_CHECKING
import time

if TYPE_CHECKING:
    from .models import Filter, FilterSettings

from . import log, VERSION
from .util import get_ip_address
from .models import FilteredRequest
from .signals import (
    request_started,
    request_finished,
)
from .options import (
    RequestFilters,
)


FilterSettingsModel: "FilterSettings" = apps.get_model(
    "request_filters",
    "FilterSettings",
    require_ready=False,
)


class RequestFilterMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        # If the request is excluded, return the response.
        # Useful for making sure the site will remain accessible for administrators.
        if RequestFilters.request_is_excluded(request):
            log(f"({self.__class__.__name__} / {get_ip_address(request)}) Request was excluded from filtering: {request.path}", log.INFO)
            return self.get_response(request)
        
        filter_start_time = time.time()
        ip_addr = get_ip_address(request)

        # Load the filters and settings from the cache.
        settings: "FilterSettings" = FilterSettingsModel.load_from_cache(request)
        filters:  list["Filter"]   = settings.get_filters()

        # Broadcast the request_started signal.
        request_started.send(sender=self, request=request, settings=settings, filters=filters)

        # Execute the filters and return the response if a filter matches,
        # otherwise return the response from get_response.
        response, passed_filters = self.execute_filters(settings, request, filters)

        # Broadcast the request_finished signal.
        request_finished.send(sender=self, request=request, settings=settings, filters=filters, response=response)

        if RequestFilters.ADD_FILTER_HEADERS:
            # Set headers to indicate that the request was filtered.
            response["X-Request-Filters"]           = "true"
            response["X-Request-Filters-Pass"]      = "true" if passed_filters else "false"
            response["X-Request-Filters-Duration"]  = str(time.time() - filter_start_time)
            response["X-Request-Filters-Version"]   = VERSION
            

        log(f"({self.__class__.__name__}) Filtering request ({ip_addr}) with {len(filters)} filters took {time.time() - filter_start_time} seconds", log.DEBUG)
        
        return response
    

    def execute_filters(self, settings: "FilterSettings", request, filters: list["Filter"]):
        """
            Execute the filters and return the response if a filter matches.
        """

        # Filter.passes_test checks if the request fits the 'shape' of the filter.
        # For example, if the type is IP, it checks if the request IP is in the filter.filter_value's CIDR range.
        # It returns true if the value is a match and thus should be filtered.
        for filter in filters:
            if filter.passes_test(settings, request):
                
                # Log and respond if filter matches.
                try:
                    response = filter.execute_action(settings, request, self.get_response)

                # Log then re-raise PermissionDenied
                except PermissionDenied as e:
                    FilteredRequest.log_request(request, filter=filter, fail_silently=True)
                    raise e
                
                # Log the request/response.
                else:
                    # Filters may return none to allow for the next filter to be executed.
                    if response is not None:
                        FilteredRequest.log_request(request, filter=filter, response=response, fail_silently=True)
                        return response, False

        # No filters matched
        response = self.get_response(request)

        if RequestFilters.LOG_HAPPY_PATH:
            FilteredRequest.log_request(request, filter=None, response=response, fail_silently=True)

        return response, True
    



import logging
from django.db import models
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _

from wagtail.admin.panels import (
    FieldPanel,
    FieldRowPanel,
)

from wagtail.models import Orderable
from modelcluster.fields import ParentalKey

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .settings import FilterSettings

from ..options import (
    RequestFilters,
)
from ..actions import (
    FilterActionChoices,
    FILTER_ACTIONS,
)
from ..checks import (
    FilterChoices,
    FilterMethodChoices,
    FILTER_CHECKS,
)
from .. import log


class FilterQuerySet(models.QuerySet):

    def active(self):
        return self.filter(active=True)
    
    def for_settings(self, settings: "FilterSettings"):
        return self.filter(settings=settings)
    
    def from_cache(self):
        filters = RequestFilters.cache_backend.get(RequestFilters.FILTERS_CACHE_KEY)
        if filters is None:
            filters = list(self)
            RequestFilters.cache_backend.set(RequestFilters.FILTERS_CACHE_KEY, filters, RequestFilters.FILTERS_CACHE_TIMEOUT.total_seconds())
        return filters

    
            
class Filter(Orderable):

    settings = ParentalKey(
        "FilterSettings",
        related_name="filters",
        on_delete=models.CASCADE,
    )

    filter_type = models.CharField(
        max_length=20,
        choices=FilterChoices.choices,
        default=FilterChoices.IP,
        verbose_name=_("Filter Type"),
    )

    method = models.CharField(
        max_length=20,
        choices=FilterMethodChoices.choices,
        default=FilterMethodChoices.ABSOLUTE,
        verbose_name=_("Method"),
    )

    action = models.CharField(
        max_length=20,
        choices=FilterActionChoices.choices,
        default=FilterActionChoices.BLOCK,
        verbose_name=_("Action"),
    )

    action_value = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_("Action Value"),
        help_text=_("The value to use for the action. For example, if you are redirecting, you would enter the URL here."),
    )

    filter_value = models.CharField(
        max_length=255,
        verbose_name=_("Filter Value"),
        help_text=_(
            "The value to filter on. For example, if you are filtering on IP, you would enter an IP address or CIDR range here."
        ),
    )

    active = models.BooleanField(
        default=True,
        verbose_name=_("Active"),
    )

    panels = [
        FieldRowPanel([
            FieldPanel("filter_type"),
            FieldPanel("method"),
            FieldPanel("action"),
            FieldPanel("active"),
        ]),
        FieldPanel("action_value"),
        FieldPanel("filter_value"),
    ]

    objects: FilterQuerySet = FilterQuerySet.as_manager()

    class Meta:
        verbose_name = _("Filter")
        verbose_name_plural = _("Filters")
        ordering = ["sort_order"]

    def __str__(self):
        s = [
            self.get_filter_type_display(),
            self.get_method_display(),
            self.get_action_display(),
            self.filter_value,
        ]
        s = " > ".join(s)
        s = s.upper()
        return f"Filter({self.pk}: {s})"

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        if RequestFilters.CLEAR_CACHE_ON_SAVE_FILTERS:
            RequestFilters.cache_backend.delete(RequestFilters.SETTINGS_CACHE_KEY)
            RequestFilters.cache_backend.delete(RequestFilters.FILTERS_CACHE_KEY)

    def passes_test(self, settings: "FilterSettings", request: HttpRequest) -> bool:
        try:
            passes_test = FILTER_CHECKS[self.method][self.filter_type]
            return passes_test(self, settings, request)
        
        except (KeyError, TypeError, ValueError):
            return RequestFilters.default_check_value(self, settings, request)

    def execute_action(self, settings: "FilterSettings", request: HttpRequest, get_response: callable):
        try:
            return FILTER_ACTIONS[self.action](self, settings, request, get_response)
        except (KeyError, TypeError, ValueError) as e:
            log(f"Error executing action on {request.path} with {self}", exc_info=e, level=logging.ERROR)
            return RequestFilters.default_action_value(self, settings, request, get_response)
        



import ipaddress
from typing import Union, TYPE_CHECKING
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from django.http import HttpRequest, HttpResponseForbidden
from django.core.exceptions import PermissionDenied
from django.core.cache import caches
from django.utils import timezone

if TYPE_CHECKING:
    from .models import (
        Filter,
        FilterSettings,
    )

import fnmatch, functools

from .util import (
    ip_in_cidr,
    get_ip_address,
)


class RequestFilterOptions(object):
    # List of excluded apps, all requests to these apps will be allowed (If resolver_match is available).
    # !Exclusions should preferably happen via IP ranges or absolute IPs.
    EXCLUDED_APPS:                list[str] = [
        "admin",
    ]

    # Excluded paths, all requests to these paths will be allowed.
    # Paths should be in the format of a glob pattern.
    # !Exclusions should preferably happen via IP ranges or absolute IPs.
    EXCLUDED_PATHS:               list[str] = [
        "/admin/*",
        f"{getattr(settings, 'STATIC_URL', '/static/')}*",
        f"{getattr(settings, 'MEDIA_URL', '/media/')}*",
    ]

    # Excluded IP addresses, all requests from these IPs will be allowed.
    # This is the safest way to exclude requests from being filtered.
    EXCLUDED_IPS:                 list[str] = [
        # "127.0.0.0/8", "::1/128",
    ]
    # Default cache backend to use for storing settings and filters
    CACHE_BACKEND:                str                   = "default"
    BLOCK_MESSAGE:                str                   = _("You are not allowed to access this resource")
    SETTINGS_CACHE_KEY:           str                   = "request_filters_settings"
    FILTERS_CACHE_KEY:            str                   = "request_filters_filters"
    SETTINGS_CACHE_TIMEOUT:       timezone.timedelta    = timezone.timedelta(minutes=5) # timezone.timedelta(seconds=5) # timezone.timedelta(minutes=5)
    FILTERS_CACHE_TIMEOUT:        timezone.timedelta    = timezone.timedelta(hours=1)   # timezone.timedelta(seconds=5) # timezone.timedelta(hours=1)
    CLEAR_CACHE_ON_SAVE_SETTINGS: bool                  = True  # Clear cache when settings are saved
    CLEAR_CACHE_ON_SAVE_FILTERS:  bool                  = True  # Clear cache when filters are saved
    DEFAULT_CHECK_VALUE:          Union[bool, callable] = True  # Allow checks to pass by default
    ADD_FILTER_HEADERS:           bool                  = True  # Add headers to the response which displays minimal information about the filters.
    LOG_HAPPY_PATH:               bool                  = False # Log requests that are allowed by the filters
    DEFAULT_ACTION_VALUE:         callable              = lambda self, filter, settings, request, get_response: HttpResponseForbidden(
        _("You are not allowed to access this resource")
    )

    def __init__(self):
        for i, ip in enumerate(self.EXCLUDED_IPS):
            if "/" in ip:
                addr = ipaddress.ip_network(ip, strict=False)
            else:
                addr = ipaddress.ip_address(ip)

            self.EXCLUDED_IPS[i] = addr


    def __getattr__(self, name):
        # Get the value from settings or the default
        return getattr(settings, name, super().__getattribute__(name))
    
    def default_check_value(self, filter: "Filter", settings: "FilterSettings", request: HttpRequest):
        """
            Default value I.E. when a check cannot be performed
        """
        if callable(self.DEFAULT_CHECK_VALUE):
            return self.DEFAULT_CHECK_VALUE(filter, settings, request)
        return self.DEFAULT_CHECK_VALUE
    
    def default_action_value(self, filter: "Filter", settings: "FilterSettings", request: HttpRequest, get_response: callable):
        """
            Default action to perform when a filter matches
        """
        return self.DEFAULT_ACTION_VALUE(filter, settings, request, get_response)

    @property
    def cache_backend(self):
        """
            Return the default cache backend for the settings and filters
        """
        return caches[self.CACHE_BACKEND]

    def request_is_excluded(self, request: HttpRequest):
        """
            Check if a request should be excluded from filtering
        """
        checks = [
            self.ip_is_excluded(
                get_ip_address(request),
            ),
            self.path_is_excluded(request.path),
        ]
        if request.resolver_match:
            checks.append(
                self.app_is_excluded(request.resolver_match.app_name)
            )

        return any(checks)
    
    def app_is_excluded(self, app_name):
        """
            Check if an app is excluded from filtering.
        """
        return app_name.lower() in self.EXCLUDED_APPS
    
    @functools.lru_cache(maxsize=128)
    def path_is_excluded(self, path):
        """
            Check if a path is excluded from filtering.
        """
        for p in self.EXCLUDED_PATHS:
            if fnmatch.fnmatch(path, p):
                return True
        return False

    @functools.lru_cache(maxsize=128)
    def ip_is_excluded(self, ip):
        """
            Check if an IP is excluded from filtering by cidr range or absolute match.
        """
        for excluded_ip in self.EXCLUDED_IPS:
            if ip_in_cidr(ip, excluded_ip):
                return True

        return False
    

# Default options for the request_filters app.
# 
# These can be overridden by settings.
# 
# Example:
#
#  REQUEST_FILTERS = {
#   "EXCLUDED_APPS": ["admin"],
#   ...
# }
#
RequestFilters = RequestFilterOptions()

del RequestFilterOptions



from typing import Literal, Union
from django.conf import settings
from django.contrib.gis.geoip2 import (
    GeoIP2,
)

import ipaddress, functools

IS_PROXIED = getattr(settings, "USE_X_FORWARDED_HOST", False)
GEO_IP = GeoIP2()


def get_ip_address(request):
    if IS_PROXIED:
        addr = request.META.get('HTTP_X_FORWARDED_FOR', None)
        if addr:
            return addr.split(',')[-1].strip()
    else:
        return request.META.get('REMOTE_ADDR', None)

def get_user_agent(request):
    return request.META.get('HTTP_USER_AGENT', "")

def get_referer(request):
    return request.META.get('HTTP_REFERER', "")

@functools.lru_cache(maxsize=128)
def ip_in_cidr(ip, cidr):
    if not isinstance(ip, ipaddress.IPv4Address) and not isinstance(ip, ipaddress.IPv6Address):
        ip = ipaddress.ip_address(ip)
        
    if isinstance(cidr, str):
        return ipaddress.ip_address(ip) in ipaddress.ip_network(cidr, strict=False)
    elif isinstance(cidr, ipaddress.IPv4Network) or isinstance(cidr, ipaddress.IPv6Network):
        return ipaddress.ip_address(ip) in cidr
    elif not (
        isinstance(cidr, ipaddress.IPv4Address) and isinstance(cidr, ipaddress.IPv6Address)
    ):
        cidr = ipaddress.ip_address(cidr)

    if ip == cidr:
        return True

    return False


@functools.lru_cache(maxsize=128)
def get_country(ip) -> dict[Union[Literal["country_code"], Literal["country_name"]], str]:
    return GEO_IP.country(ip)

def skip_logging(request_or_response):
    """
        Mark a response or request.
        It will not be logged to the database.
    """
    setattr(request_or_response, "filters_skip_logging", True)
    return request_or_response

def logging_skipped(request_or_response):
    """
        Check if a request or response has been marked as logging skipped.
    """
    return hasattr(request_or_response, "filters_skip_logging") and request_or_response.skip_logging




import logging
from django.db import models
from django.conf import settings as django_settings
from django.core.exceptions import PermissionDenied
from django.http import (
    HttpRequest,
    HttpResponseRedirect,
    JsonResponse,
)
from django.utils.translation import gettext_lazy as _
from django.utils.translation.trans_real import language_code_prefix_re
from django.utils.translation import get_language_from_path
from typing import TYPE_CHECKING

from . import log
from .util import (
    get_ip_address,
    skip_logging,
)
from .signals import (
    request_blocked,
    request_redirected,
)
from .options import (
    RequestFilters,
)

if TYPE_CHECKING:
    from .models import Filter, FilterSettings


def strip_language_from_path(path):
    # Get the language code from the path
    language_code = get_language_from_path(path)
    if language_code:
        # If the language code is found, remove it from the path
        path = language_code_prefix_re.sub("", path, 1)
    return path



class FilterActionChoices(models.TextChoices):
    ALLOW       = "ALLOW",      _("Allow")
    BLOCK       = "BLOCK",      _("Block")
    REDIRECT    = "REDIRECT",   _("Redirect")
    LOG         = "LOG",        _("Log")


def FILTER_ALLOWS_REQUEST(filter: "Filter", settings: "FilterSettings", request: HttpRequest, get_response):
    """
        Lets a request pass - skips any further filters and their checks.
        Filters in the list before this action will still be checked and acted upon.
    """
    ipaddr = get_ip_address(request)
    log(f"[Filter / {filter.action}] Allowing request {ipaddr} with {filter}", logging.INFO)
    return get_response(request)

def FILTER_BLOCKS_REQUEST(filter: "Filter", settings: "FilterSettings", request: HttpRequest, get_response):
    """
        Blocks a request - does not execute any further filters or their checks.
    """

    # Requests don't get blocked for no reason - log with a reasonably high level.
    ipaddr = get_ip_address(request)
    log(f"[Filter / {filter.action}] Blocking request {ipaddr} with {filter}", logging.WARN)

    # Broadcast the request_blocked signal
    request_blocked.send(sender=filter, request=request, filter=filter)

    if request.accepts("application/json"):
        return JsonResponse({
            "error": RequestFilters.BLOCK_MESSAGE,
        }, status=403)

    raise PermissionDenied(RequestFilters.BLOCK_MESSAGE)
 

def FILTER_REDIRECTS_REQUEST(filter: "Filter", settings: "FilterSettings", request: HttpRequest, get_response):
    to = filter.action_value or "/"
    to = to.strip()
    
    ipaddr = get_ip_address(request)

    response = None
    if django_settings.USE_I18N:
        # If the path has a language prefix, strip it from the path
        cmp = strip_language_from_path(request.path)
        cmp = cmp.strip()
        if cmp == "":
            cmp = "/"

        if cmp and to == cmp:
            response = get_response(request)
    
    else:
        # If the path is the same as the redirect, return the response
        if to == request.path:
            response = get_response(request)
    

    if response is None:
        # Unsure about stability of redirects here - log INFO.
        log(f"[Filter / {filter.action}] Redirecting request {ipaddr} with {filter} from {request.path} to {to}", logging.INFO)
        response = HttpResponseRedirect(to)
    else:
        log(f"[Filter / {filter.action}] Request {ipaddr} with {filter} does not need redirection from {request.path} to {to}!", logging.DEBUG)

    # Broadcast the request_redirected signal
    request_redirected.send(sender=filter, request=request, filter=filter, response=response)

    return response

import json, logging
from urllib.parse import urlencode

def _FILTER_LOGS_REQUEST(log_level: int, allow_next_filter: bool = False):
    def _filter_logs_request(filter: "Filter", settings: "FilterSettings", request: HttpRequest, get_response):
        # Get the response and skip the logging in middleware 
        # to implement our custom logging here.
        response = get_response(request)
        response = skip_logging(response)
        _content_length = request.META.get('CONTENT_LENGTH', 0) or 0
        try:
            if int(_content_length) > 0:
                body = request.body
            else:
                body = None
        except (TypeError, ValueError):
            body = None

        ip_addr = get_ip_address(request)
        method = request.method
        path = request.path
        query = request.GET
        status_code = response.status_code

        meta = {k: v for k, v in request.headers.items()}
        meta_str = json.dumps(meta, indent=2, sort_keys=True)
        message = f"[Filter / {filter.action} / {ip_addr} / {method}: {status_code}] {path} {urlencode(query)}\nMETA:\n{meta_str}\n\nBODY:\n{body}\n"

        log(message, level=log_level)

        if allow_next_filter:
            return None
        
        return response

    return _filter_logs_request


FILTER_LOGS_REQUEST = _FILTER_LOGS_REQUEST(logging.DEBUG, False)
FILTER_LOGS_REQUEST_ALLOW_NEXT = _FILTER_LOGS_REQUEST(logging.DEBUG, True)


FILTER_ACTIONS = {
    FilterActionChoices.ALLOW:      FILTER_ALLOWS_REQUEST,
    FilterActionChoices.BLOCK:      FILTER_BLOCKS_REQUEST,
    FilterActionChoices.REDIRECT:   FILTER_REDIRECTS_REQUEST,
    FilterActionChoices.LOG:        FILTER_LOGS_REQUEST_ALLOW_NEXT,
}

